// Ethernet Watchdog Board Firmware
// V3.0 -- Version for Ethernet Watchdog PCB
//		   with Serial/USB watchdog support
//
// 08/04/2020
// Randall Hyde
//
// Copyright 2020, Plantation Productions, Inc.

#include <SPI.h>
#include <Ethernet.h>

/*
  INLWatchdog.ino (Teensy 3.2 version on the Ethernet WD PCB)
  Maintains watchdog timers on Ethernet ports 20560 and 20561
  or on Serial input port, based on the value of Pin 2
*/


#define useSerialPort 0
#define TRUE 1

// Timeout period for the watchdog timers (in 1/2 second counts)

#define timeout1 (14)
#define timeout2 (14)

// Pin 7 is the Heartbeat LED on the Ethernet WD board
// give it a name:

int led = (7);

// Pin 2 determines the watchdog mode. 1=USB/Serial, 0=Ethernet

#define wd_mode     (2)


#define eth_reset	(9)
#define wd_rfsh     (22)
#define wd_latch    (21)
#define Reset       (23)


// Eight output pins connected to ULN2803 Darlington array.
// Note: Pins are reversed to improve the PCB layout.
// Relay outputs:

#define out7 (24)
#define out6 (25)
#define out5 (26)
#define out4 (27)
#define out3 (28)
#define out2 (29)
#define out1 (30)
#define out0 (31)


// Ethernet MAC address.
// Currently "PPEWD2" (for Plantation Productions
// Ethernet Watch Dog Timer, Rev 2).

byte mac[] = { 'P', 'P', 'E', 'W', 'D', '2' };

// IP Address for watchdog timer:

IPAddress ip( 192,168,2,19);

// Timer1Port, Timer2Port-
// The two socket ports we listen to in order
// to update the two watchdog timers.
// Note: 20560 is the standard Plantation
// Productions port (0x5050, or "PP).

#define timer1Port (20560)
#define timer2Port (20561)


// Ethernet WD Board + Robotdyn Ethernet interface
// uses Teensy3.2 chip select 10:

#define WIZ_CS (10)

EthernetServer server1(timer1Port);
EthernetServer server2(timer2Port);

// W5500 chip (Robotdyn Ethernet IF) supports up
// to 8 sockets. Let each client have four of them.

#define maxClients (4)

EthernetClient clients1[maxClients];
EthernetClient clients2[maxClients];


// wdMode- 1 if Serial mode, 0 if Ethernet mode


enum wdMode_t
{
	ethernetMode,
	serialMode
};

wdMode_t wdMode = ethernetMode;


// Turn off all the PWM outputs:

void clearAll( void )
{
    	digitalWrite( out0, 0 );
		digitalWrite( out1, 0 );
		digitalWrite( out2, 0 );
		digitalWrite( out3, 0 );
		digitalWrite( out4, 0 );
		digitalWrite( out5, 0 );
		digitalWrite( out6, 0 );
		digitalWrite( out7, 0 );
}


// Must send a positive pulse to the TA8030s
// to refresh the watchdog timer:

void watchdogRefresh( void )
{
    digitalWrite( wd_rfsh, 1 );
    delay( 1 );
    digitalWrite( wd_rfsh, 0 );
}

// Must send a negative-going pulse to the TA8030S
// chip to cause a reset. The pulse must be at least
// 350 msec wide.

void toggleReset( void )
{
        digitalWrite( Reset, 1 );
        delay( 500 );
        digitalWrite( Reset, 0 );
        delay( 500 );
        digitalWrite( Reset, 1 );
}

// Must send a negative-going pulse to the W5500
// chip to cause a reset. Make the pulse 1ms wide.

void toggleEthReset( void )
{
        digitalWrite( eth_reset, 1 );
        delayMicroseconds( 10 );
        digitalWrite( eth_reset, 0 );
        delayMicroseconds( 10 );
        digitalWrite( eth_reset, 1 );
        delayMicroseconds( 10 );
}

// the setup routine runs once when you press reset:

void setup()
{
	byte latch;



	// Initialize the wd_mode pin as an input and
	// turn on the pullup resistor:

	pinMode(wd_mode, INPUT_PULLUP);


    // Initialize the LED digital pin as an output.

    pinMode(led, OUTPUT);

    // PWM output lines -- always outputs:

    pinMode( out0, OUTPUT );
    pinMode( out1, OUTPUT );
    pinMode( out2, OUTPUT );
    pinMode( out3, OUTPUT );
    pinMode( out4, OUTPUT );
    pinMode( out5, OUTPUT );
    pinMode( out6, OUTPUT );
    pinMode( out7, OUTPUT );

	clearAll();					// Assume time-out (failsafe) on startup.

    // Stuff needed by the watchdog timer:

    pinMode( eth_reset, OUTPUT );   // Ethernet reset pin is an output
    pinMode( Reset, OUTPUT );       // Reset pin is an output
    pinMode( wd_rfsh, OUTPUT );     // Watchdog refresh is an output
    pinMode( wd_latch, INPUT );     // Watchdog Latch is an input

	wdMode = (digitalRead( wd_mode ) == 0) ? ethernetMode : serialMode;

    // Initialize the Ethernet featherwing:


	#if useSerialPort

		Serial.begin( 9600 );
		while(!Serial);	// Wait for serial port to come up.
	    Serial.println( "Plantation Productions' Ethernet Watchdog timer" );

	#else

		if( wdMode == serialMode )
		{
			Serial.begin( 9600 );
			while(!Serial);	// Wait for serial port to come up.
		}

	#endif

    // Loop until we've reset the watchdog timer and read a "watchdog latch
    // not set" signal from the watchdog timer.

    do
    {
    	toggleEthReset();				// Reset the ethernet chip, too.
		toggleReset();                  // Reset watchdog timer latch.
		watchdogRefresh();              // Must do this at least once every 4.7 seconds
		latch = digitalRead( wd_latch );
		#if useSerialPort

			if( !latch )
			{
				Serial.println( "Watchdog latch not cleared, resetting" );
			}

		#endif

	}while( !latch );


	// Determine if we're running in Serial mode or Ethernet mode

	#if useSerialPort

		Serial.print( "wd_mode=" );
		int i = digitalRead( wd_mode );
		Serial.println( i );
		Serial.print( "wdMode = " );
		Serial.println( wdMode );

	#endif

	if( wdMode == ethernetMode )
	{

		// Initialize the Ethernet hardware
		// and make sure it comes up:

		do
		{
			toggleEthReset();				// Reset the ethernet chip, too.
			delay( 100 );
			Ethernet.init(WIZ_CS);
			Ethernet.begin(mac, ip);

			while(Ethernet.hardwareStatus() == EthernetNoHardware)
			{
				#if useSerialPort

					Serial.println("Ethernet shield was not found.  Resetting...");

				#endif
				toggleEthReset();
				delay(10);
				Ethernet.init(WIZ_CS);
				Ethernet.begin(mac, ip);
			}
			#if useSerialPort


				Serial.print( "Ethernet Watchdog Timer IP address: ");
				Serial.println( Ethernet.localIP());
				Serial.print( "Ethernet Watchdog port: ");
				Serial.println( timer1Port );

			#endif
		}
		while( Ethernet.localIP() != ip );

		server1.begin();
		server2.begin();

	} // endif wdMode



}

// the loop routine runs over and over again forever:

void loop()
{
	static unsigned long lastTime = 0;
	static unsigned long timer1 = 0;
	static unsigned long timer2 = 0;
	unsigned long curTime = millis();

	static byte ledState = 0;
	static byte WD1State = 0;
	static byte WD2State = 0;



	// Check port timer1Port.
	// If we receive a character on timer1Port,
	// then reset the timer1 time-out timer.

	if( wdMode == ethernetMode )
	{
		EthernetClient newClient1 = server1.accept();
		if (newClient1)
		{
			// Find an empty slot on clients1 and put
			// the new client in that slot:

			for (byte i = 0; i < maxClients; i++)
			{
				if (!clients1[i]) // If there is not a client in slot "i".
				{
					// Once we "accept", the client is no longer tracked by EthernetServer
					// so we must store it into our list of clients

					clients1[i]			= newClient1;
					#if useSerialPort

						Serial.print( "Connecting to client 1 in slot1:" );
						Serial.println( i );

					#endif
					break;
				}
			}
		}

		// check for incoming data from all clients of server 1

		for (byte i = 0; i < maxClients; i++)
		{
			if( clients1[i] ) // If there is a client in slot "i".
			{
				while( clients1[i].available() > 0 ) // While chars available
				{
					// read incoming data from the client

					char c = clients1[i].read();

					#if useSerialPort

						Serial.print( "Received character '" );
						Serial.print(c);
						Serial.print( "' from client " );
						Serial.println( i );

					#endif
					timer1 = 0;	// Reset watchdog timer #1
				}
			}
		}

		// Stop any clients of server1 which disconnect

		for (byte i = 0; i < maxClients; i++)
		{
			if (clients1[i] && !clients1[i].connected())
			{
				#if useSerialPort

					Serial.print( "Disconnecting from client 1 in slot1:" );
					Serial.print( i );
					Serial.print( ", socket #" );
					Serial.print( clients1[i].getSocketNumber());
					Serial.print( " (max=");
					Serial.print( MAX_SOCK_NUM );
					Serial.println( ")" );

				#endif
				clients1[i].stop();
			}
		}



		// Check port timer2Port.
		// If we receive a character on timer2Port,
		// then reset the timer2 time-out timer.


		EthernetClient newClient2 = server2.accept();
		if (newClient2)
		{
			// Find an empty slot on clients2 and put
			// the new client in that slot:

			for (byte i = 0; i < maxClients; i++)
			{
				if (!clients2[i]) // If there is not a client in slot "i".
				{
					// Once we "accept", the client is no longer tracked by EthernetServer
					// so we must store it into our list of clients

					clients2[i] = newClient2;
					#if useSerialPort

						Serial.print( "Connecting to client 2 in slot2:" );
						Serial.println( i );

					#endif
					break;
				}
			}
		}

		// check for incoming data from all clients of server 2

		for (byte i = 0; i < maxClients; i++)
		{
			if( clients2[i] ) // If there is a client in slot "i".
			{
				while( clients2[i].available() > 0 ) // While chars available
				{
					// read incoming data from the client

					char c = clients2[i].read();
					timer2 = 0;	// Reset watchdog timer #1
				}
			}
		}

		// Stop any clients of server2 which disconnect

		for (byte i = 0; i < maxClients; i++)
		{
			if (clients2[i] && !clients2[i].connected())
			{
				#if useSerialPort

					Serial.print( "Disconnecting from client 2 in slot2:" );
					Serial.print( i );
					Serial.print( ", socket #" );
					Serial.print( clients2[i].getSocketNumber());
					Serial.print( " (max=");
					Serial.print( MAX_SOCK_NUM );
					Serial.println( ")" );

				#endif
				clients2[i].stop();
			}
		}

	} // wdMode == ethernetMode
	else // wdMode == serialMode
	{
		if( Serial.available() > 0 )
		{
			char c = Serial.read();
			#if useSerialPort

				Serial.print( "Received character '" );
				Serial.println(c);

			#endif
			timer1 = 0;	// Reset watchdog timer #1


		}
	}





	// Every 1/2 second invert the LED and
	// refresh the watchdog timer.
	//
	// Note that millis() will roll-over to
	// zero every 50 days or so. When that happens
	// the following will produce a value that
	// is definitely greater than 500. However,
	// this just (worst case) reduces the time
	// out period by, at most, 1/2 second (typically
	// far less). This won't be an issue at all.

	if( abs( curTime - lastTime ) >= 500 )
	{

    	digitalWrite(led, ledState);
    	ledState = !ledState;
    	watchdogRefresh();
    	lastTime = curTime;

    	// If the Weensy watchdog timer has timed out,
    	// reset it (because we're obviously running).

    	byte latch = digitalRead( wd_latch );
    	if( !latch )
    	{
    		#if useSerialPort

				Serial.println( "Watchdog latch clear, resetting" );

			#endif

			// Reset the Ethernet clients:

			if( wdMode == ethernetMode )
			{
				for( byte i=0; i < maxClients; ++i )
				{
					if( clients1[i] )
					{
						clients1[i].stop();
						#if useSerialPort

							Serial.print( "Disconnecting from client 1 in slot1:" );
							Serial.print( i );
							Serial.print( ", socket #" );
							Serial.print( clients1[i].getSocketNumber());
							Serial.print( " (max=");
							Serial.print( MAX_SOCK_NUM );
							Serial.println( ")" );

						#endif
					}
					if( clients2[i] )
					{
						#if useSerialPort

							Serial.print( "Disconnecting from client 2 in slot2:" );
							Serial.print( i );
							Serial.print( ", socket #" );
							Serial.print( clients2[i].getSocketNumber());
							Serial.print( " (max=");
							Serial.print( MAX_SOCK_NUM );
							Serial.println( ")" );

						#endif
						clients2[i].stop();
					}
				}
			} // endif wdMode
			setup();	// Reset everything
			timer1 = 0;
			timer2 = 0;
		}

    	// Bump up the watchdog timers every 1/2 second

    	++timer1;
    	++timer2;

    }

    // Test to see if the Watchdog #1 timer has timed out:


    if( timer1 < timeout1 )	// 14 * 0.5s = 7 seconds
    {

    	// No time out, so set the relays high.
    	// This will turn the relays back on if
    	// they've been off because of a watchdog timeout.

    	digitalWrite( out0,  HIGH );	// No watchdog timeout.
		digitalWrite( out3,  HIGH );	// Controls CH1 LED
    	if( wdMode == ethernetMode )
    	{
			digitalWrite( out1,  HIGH );
			digitalWrite( out2,  HIGH );
		}
		WD1State = 0;
	}
	else
	{
		// If watchdog timer #1 has timed out (7 seconds),
		// then turn the relays off.

    	digitalWrite( out0,  LOW );	// Watchdog timeout
		digitalWrite( out1,  LOW );
		digitalWrite( out2,  LOW );
		digitalWrite( out3,  LOW );
		#if useSerialPort

			if( !WD1State )
			{
				WD1State = 1;
				Serial.println( "Watchdog 1 timed out" );
			}

		#endif
	}

	// See if the Watchdog #2 timer has timed out.

	if( wdMode == ethernetMode )
	{
		if( timer2 < timeout2 )	// 14 * 0.5s = 7 seconds
		{
			digitalWrite( out4,  HIGH );	// No watchdog timeout.
			digitalWrite( out5,  HIGH );
			digitalWrite( out6,  HIGH );
			digitalWrite( out7,  HIGH );
			WD2State = 0;
		}
		else
		{
			digitalWrite( out4,  LOW );	// Watchdog timeout
			digitalWrite( out5,  LOW );
			digitalWrite( out6,  LOW );
			digitalWrite( out7,  LOW );
			#if useSerialPort

				if( !WD2State )
				{
					WD2State = 1;
					Serial.println( "Watchdog 2 timed out" );
				}

			#endif
		}
	} //endif wdMode


}
